"""Set up a fake hook for testing purposes."""

import platform
from pathlib import Path
from typing import Dict

import pluggy
import pytest

from edgetest import hookspecs

hookimpl = pluggy.HookimplMarker("edgetest")


class FakeHook:
    """Create a series of fake hooks."""

    @hookimpl
    def path_to_python(self, basedir: str, envname: str) -> str:
        """Return the path to the python executable."""
        # return str(Path(basedir, envname, "bin", "python"))
        if platform.system() == "Windows":
            return str(Path(basedir) / envname / "Scripts" / "python")
        else:
            return str(Path(basedir) / envname / "bin" / "python")

    @hookimpl
    def create_environment(self, basedir: str, envname: str, conf: Dict):
        """Create the virtual environment for testing."""
        pass


@pytest.fixture
def plugin_manager():
    """The plugin manager for our fake hook."""
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.register(FakeHook())

    return pm



"""Default virtual environment hook."""

import platform
from pathlib import Path
from typing import Dict
from venv import EnvBuilder

import pluggy

hookimpl = pluggy.HookimplMarker("edgetest")


@hookimpl(trylast=True)
def path_to_python(basedir: str, envname: str) -> str:
    """Return the path to the python executable."""
    if platform.system() == "Windows":
        return str(Path(basedir) / envname / "Scripts" / "python")
    else:
        return str(Path(basedir) / envname / "bin" / "python")


@hookimpl(trylast=True)
def create_environment(basedir: str, envname: str, conf: Dict):
    """Create the virtual environment for testing.

    Creates an environment using ``venv``.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        The name of the virtual environment.
    conf : dict
        Ignored.

    Raises
    ------
    RuntimeError
        Error raised if the environment cannot be created.
    """
    builder = EnvBuilder(with_pip=True)
    try:
        builder.create(env_dir=Path(basedir, envname))
    except Exception:
        raise RuntimeError(f"Unable to create {envname} in {basedir}")



"""Utility functions."""

import os
from configparser import ConfigParser
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Dict, List, Optional, Tuple

from packaging.specifiers import SpecifierSet
from pkg_resources import parse_requirements

from .logger import get_logger

LOG = get_logger(__name__)


def _run_command(*args) -> Tuple[str, int]:
    """Run a command using ``subprocess.Popen``.

    Parameters
    ----------
    *args
        Arguments for the command.

    Returns
    -------
    str
        The output
    int
        The exit code

    Raises
    ------
    RuntimeError
        Error raised when the command is not successfully executed.
    """
    LOG.debug(f"Running the following command: \n\n {' '.join(args)}")
    popen = Popen(args, stdout=PIPE, universal_newlines=True)
    out, _ = popen.communicate()
    if popen.returncode:
        raise RuntimeError(
            f"Unable to run the following command: \n\n {' '.join(args)}"
        )

    return out, popen.returncode


@contextmanager
def pushd(new_dir: str):
    """Create a context manager for running commands in sub-directories.

    Parameters
    ----------
    new_dir : str
        The relative directory to run the command in.
    """
    curr_dir = Path.cwd()
    os.chdir(curr_dir / new_dir)
    try:
        yield
    finally:
        os.chdir(curr_dir)


def convert_requirements(requirements: str, conf: Optional[Dict] = None) -> Dict:
    """Generate environments for a newline-separate list of package requirements.

    This function will generate one environment per entry with an additional environment
    that upgrades all requirements simultaneously.

    Parameters
    ----------
    requirements : str
        The requirements string.
    conf : dict, optional (default None)
        An existing configuration to edit.

    Returns
    -------
    Dict
        A configuration dictionary.
    """
    conf = {"envs": []} if conf is None else conf
    pkgs = [pkg.project_name for pkg in parse_requirements(requirements)]
    for pkg in pkgs:
        conf["envs"].append({})
        conf["envs"][-1]["name"] = pkg
        conf["envs"][-1]["upgrade"] = pkg
    # Create an environment with all requirements upgraded
    conf["envs"].append({})
    conf["envs"][-1]["name"] = "all-requirements"
    conf["envs"][-1]["upgrade"] = "\n".join(pkgs)

    return conf


def gen_requirements_config(fname_or_buf: str, **options) -> Dict:
    """Generate a configuration file from package requirements.

    This function will convert the package installation requirements to a configuration
    file with one environment per requirement.

    Parameters
    ----------
    fname_or_buf : str
        Path to the requirements file to parse using ``pkg_resources.parse_requirements``
        or the string representing the requirements file.
    **options
        Options to apply to each test environment.

    Returns
    -------
    Dict
        The configuration file.
    """
    # First, get the requirements
    if Path(fname_or_buf).is_file():
        with open(fname_or_buf) as infile:
            cfg = infile.read()
    else:
        cfg = fname_or_buf
    output = convert_requirements(requirements=cfg)
    for index in range(len(output["envs"])):
        output["envs"][index].update(options)

    return output


def parse_cfg(filename: str = "setup.cfg", requirements: Optional[str] = None) -> Dict:
    """Generate a configuration from a ``.ini`` style file.

    This function can operate in two ways. First, it can look for sections that
    start with ``edgetest`` and build a configuration. Suppose
    you have ``setup.cfg`` as follows:

    .. code-block:: ini

        [edgetest.envs.pandas]
        upgrade =
            pandas

    This will result in a configuration that has one testing environment, named
    ``pandas``, that upgrades the ``pandas`` package.

    If you don't have any sections that start with ``edgetest.envs``, we will look for
    the PEP 517-style ``setup.cfg`` install requirements (the ``install_requires`` key
    within the ``options`` section). To set global defaults for you environments, use
    the ``edgetest`` section:

    .. code-block:: ini

        [edgetest]
        extras =
            tests
        command =
            pytest tests -m "not integration"

        [edgetest.envs.pandas]
        upgrade =
            pandas

    For this single environment file, the above configuration is equivalent to

    .. code-block:: ini

        [edgetest.envs.pandas]
        extras =
            tests
        command =
            pytest tests -m "not integration"
        upgrade =
            pandas

    Parameters
    ----------
    filename : str, optional (default "setup.cfg")
        The name of the configuration file to read. Defaults to ``setup.cfg``.
    requirements : str, optional (default None)
        An optional path to the requirements text file. If there are no PEP-517
        style dependencies or coded environments in the edgetest configuration, this
        function will look for dependencies in the requirements file.

    Returns
    -------
    Dict
        A configuration dictionary for ``edgetest``.
    """
    # Read in the configuration file
    config = ConfigParser()
    config.read(filename)
    # Parse
    output: Dict = {"envs": []}
    # Get any global options if necessary
    options = dict(config["edgetest"]) if "edgetest" in config else {}
    # Next, create the sections
    for section in config.sections():
        if not section.startswith("edgetest."):
            continue
        # Look for the special ``envs`` key
        section_name = section.split(".")
        if section_name[1] == "envs":
            output["envs"].append(dict(config[section]))
            output["envs"][-1]["name"] = section_name[2]
        else:
            output[section_name[1]] = dict(config[section])
    if len(output["envs"]) == 0:
        if config.get("options", "install_requires"):
            output = convert_requirements(
                requirements=config["options"]["install_requires"], conf=output
            )
        elif requirements:
            req_conf = gen_requirements_config(fname_or_buf=requirements)
            output["envs"] = req_conf["envs"]
        else:
            raise ValueError("Please supply a valid list of environments to create.")
    # Apply global environment options (without overwriting)
    for idx in range(len(output["envs"])):
        output["envs"][idx] = dict(
            list(options.items()) + list(output["envs"][idx].items())
        )

    return output


def upgrade_requirements(
    fname_or_buf: str, upgraded_packages: List[Dict[str, str]]
) -> str:
    """Create an upgraded requirements file.

    Parameters
    ----------
    fname_or_buf : str
        Path to the requirements file to parse using ``pkg_resources.parse_requirements``
        or the string representing the requirements file.
    upgraded_packages : list
        A list of packages upgraded in the testing procedure.

    Returns
    -------
    str
        The string file representing the new requirements file.
    """
    # Get the existing file
    try:
        if Path(fname_or_buf).is_file():
            with open(fname_or_buf) as infile:
                cfg = infile.read()
        else:
            cfg = fname_or_buf
    except OSError:
        # Filename too long for the is_file() function
        cfg = fname_or_buf
    pkgs = [pkg for pkg in parse_requirements(cfg)]
    upgrades = {pkg["name"]: pkg["version"] for pkg in upgraded_packages}

    for pkg in pkgs:
        if pkg.project_name not in upgrades:
            continue
        # Replace the spec
        specs = deepcopy(pkg.specs)
        for index, value in enumerate(specs):
            if value[0] == "<=":
                pkg.specs[index] = ("<=", upgrades[pkg.project_name])
            elif value[0] == "<":
                pkg.specs[index] = ("!=", value[1])
                pkg.specs.append(("<=", upgrades[pkg.project_name]))
            elif value[0] == "==":
                pkg.specs = [(">=", value[1]), ("<=", upgrades[pkg.project_name])]
        pkg.specifier = SpecifierSet(",".join("".join(spec) for spec in pkg.specs))  # type: ignore

    return "\n".join(str(pkg) for pkg in pkgs)


def upgrade_setup_cfg(
    upgraded_packages: List[Dict[str, str]], filename: str = "setup.cfg"
) -> ConfigParser:
    """Upgrade the ``setup.cfg`` file.

    Parameters
    ----------
    upgraded_packages : List[Dict[str, str]]
        A list of packages upgraded in the testing procedure.
    filename : str, optional (default "setup.cfg")
        The name of the configuration file to read. Defaults to ``setup.cfg``.

    Returns
    -------
    ConfigParser
        The updated configuration file.
    """
    parser = ConfigParser()
    parser.read(filename)
    if "options" in parser and parser.get("options", "install_requires"):
        LOG.info(f"Updating the requirements in {filename}")
        upgraded = upgrade_requirements(
            fname_or_buf=parser["options"]["install_requires"].lstrip(),
            upgraded_packages=upgraded_packages,
        )
        parser["options"]["install_requires"] = "\n" + upgraded
    # Update the extras, if necessary
    if "options.extras_require" in parser:
        for extra, dependencies in parser.items("options.extras_require"):
            upgraded = upgrade_requirements(
                fname_or_buf=dependencies,
                upgraded_packages=upgraded_packages,
            )
            parser["options.extras_require"][extra] = "\n" + upgraded

    return parser



"""Test utility functions."""

from pathlib import Path
from unittest.mock import mock_open, patch

from edgetest.schema import BASE_SCHEMA, EdgetestValidator, Schema
from edgetest.utils import gen_requirements_config, parse_cfg

REQS = """
mydep1>=0.1.0,<=0.2.0
mydep2<=0.10.0
"""

CFG_NOREQS = """
[edgetest.envs.myenv]
upgrade =
    myupgrade
command =
    pytest tests -m 'not integration'
"""

CFG_REQS = """
[options]
install_requires =
    myupgrade
"""

CFG_REQS_DEFAULTS = """
[options]
install_requires =
    myupgrade

[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'
"""

CFG_NOREQS_DEFAULTS = """
[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'

[edgetest.envs.myenv]
upgrade =
    myupgrade
command =
    pytest tests
"""

CFG_CUSTOM = """
[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'

[edgetest.custom]
mycustom = mykey

[edgetest.envs.myenv]
upgrade =
    myupgrade
"""


@patch("edgetest.utils.Path")
def test_parse_reqs(mock_pathlib):
    """Test creating a configuration from requirements."""
    mock_pathlib.return_value.is_file.return_value = True
    with patch("edgetest.utils.open", mock_open(read_data=REQS)):
        cfg = gen_requirements_config("filename")

    assert cfg == {
        "envs": [
            {"name": "mydep1", "upgrade": "mydep1"},
            {"name": "mydep2", "upgrade": "mydep2"},
            {"name": "all-requirements", "upgrade": "mydep1\nmydep2"},
        ]
    }
    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg(tmpdir):
    """Test parsing a config with no install requirements."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.ini")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_NOREQS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "command": "\npytest tests -m 'not integration'",
            }
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_default(tmpdir):
    """Test parsing a config with no install requirements and defaults."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.ini")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_NOREQS_DEFAULTS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "extras": "\ntests",
                "command": "\npytest tests",
            }
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_reqs(tmpdir):
    """Test parsing a PEP-517 style config."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_REQS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {"name": "myupgrade", "upgrade": "myupgrade"},
            {"name": "all-requirements", "upgrade": "myupgrade"},
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_reqs_default(tmpdir):
    """Test parsing a PEP-517 style config with default arguments."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_REQS_DEFAULTS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myupgrade",
                "upgrade": "myupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
            {
                "name": "all-requirements",
                "upgrade": "myupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_custom_cfg(tmpdir):
    """Test parsing a custom configuration."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_CUSTOM)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "custom": {"mycustom": "mykey"},
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            }
        ],
    }

    schema = Schema()
    schema.add_globaloption(
        "custom", {"type": "dict", "schema": {"mycustom": {"type": "string"}}}
    )

    validator = EdgetestValidator(schema=schema.schema)

    assert validator.validate(cfg)



"""Command-line interface."""

from pathlib import Path
from typing import List

import click
import pluggy

from . import hookspecs, lib
from .core import TestPackage
from .logger import get_logger
from .report import gen_report
from .schema import EdgetestValidator, Schema
from .utils import (
    gen_requirements_config,
    parse_cfg,
    upgrade_requirements,
    upgrade_setup_cfg,
)

LOG = get_logger(__name__)


def get_plugin_manager() -> pluggy.PluginManager:
    """Get the plugin manager.

    Registers the default ``venv`` plugin.

    Returns
    -------
    PluginManager
        The plugin manager.
    """
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.load_setuptools_entrypoints("edgetest")
    pm.register(lib)

    return pm


@click.command()
@click.option(
    "--config",
    "-c",
    default=None,
    type=click.Path(exists=True),
    help="Path to the test configuration file",
)
@click.option(
    "--requirements",
    "-r",
    default="requirements.txt",
    type=click.Path(),
    help="Path to a requirements file",
)
@click.option(
    "--environment",
    "-e",
    default=None,
    help="Name of a specific environment to run",
)
@click.option(
    "--notest",
    is_flag=True,
    help="Whether or not to run the test command for each environment",
)
@click.option(
    "--nosetup",
    is_flag=True,
    help="Whether or not to only set up the conda environment(s)",
)
@click.option(
    "--extras",
    type=str,
    multiple=True,
    default=None,
    help="List of extra installations for the local package. Only used if using ``requirements``",
)
@click.option(
    "--deps",
    "-d",
    type=str,
    multiple=True,
    default=None,
    help="Additional `pip` dependencies to install. Only used if using ``requirements``.",
)
@click.option(
    "--command",
    type=str,
    default="pytest",
    help="The test command to use in each environment. Only used if using ``requirements``.",
)
@click.option(
    "--export",
    is_flag=True,
    help="Whether or not to export the updated requirements file. Overwrites input requirements.",
)
def cli(
    config,
    requirements,
    environment,
    notest,
    nosetup,
    extras,
    deps,
    command,
    export,
):
    """Create the environments and test.

    If you do not supply a YAML configuration file, this package will search for a
    ``requirements.txt`` file and create a conda environment for each package in that file.
    """
    # Get the hooks
    pm = get_plugin_manager()
    if config:
        conf = parse_cfg(filename=config, requirements=requirements)
    else:
        # Find the path to the local directory using the requirements file
        conf = gen_requirements_config(
            fname_or_buf=requirements,
            extras=extras,
            deps=deps,
            command=command,
            package_dir=str(Path(requirements).parent),
        )

    # Validate the configuration file
    docstructure = Schema()
    pm.hook.addoption(schema=docstructure)
    validator = EdgetestValidator(schema=docstructure.schema)
    if not validator.validate(conf):
        click.echo("Unable to validate configuration file.")
        raise ValueError("Unable to validate configuration file.")
    conf = validator.document

    if environment:
        conf["envs"] = [env for env in conf["envs"] if env["name"] == environment]

    # Run the pre-test hook
    pm.hook.pre_run_hook(conf=conf)
    testers: List[TestPackage] = []
    for env in conf["envs"]:
        testers.append(
            TestPackage(
                hook=pm.hook,
                envname=env["name"],
                upgrade=env["upgrade"],
                package_dir=env["package_dir"],
            )
        )
        # Set up the test environment
        if nosetup:
            click.echo(f"Using existing environment for {env['name']}...")
        else:
            testers[-1].setup(**env)
        # Run the tests
        if notest:
            click.echo(f"Skipping tests for {env['name']}")
        else:
            testers[-1].run_tests(env["command"])

    report = gen_report(testers)
    click.echo(f"\n\n{report}")

    if export and testers[-1].status:
        if config is not None and Path(config).name == "setup.cfg":
            parser = upgrade_setup_cfg(
                upgraded_packages=testers[-1].upgraded_packages(),
                filename=config,
            )
            with open(config, "w") as outfile:
                parser.write(outfile)
            if "options" not in parser or not parser.get("options", "install_requires"):
                click.echo(
                    "No PEP-517 style requirements in ``setup.cfg`` to update. Updating "
                    f"{requirements}"
                )
                upgraded = upgrade_requirements(
                    fname_or_buf=requirements,
                    upgraded_packages=testers[-1].upgraded_packages(),
                )
                with open(requirements, "w") as outfile:
                    outfile.write(upgraded)
        else:
            click.echo(f"Overwriting the requirements file {requirements}...")
            upgraded = upgrade_requirements(
                fname_or_buf=requirements,
                upgraded_packages=testers[-1].upgraded_packages(),
            )
            with open(requirements, "w") as outfile:
                outfile.write(upgraded)

    # Run the post-test hook
    pm.hook.post_run_hook(testers=testers, conf=conf)



"""Create a fake package and test."""

from pathlib import Path

import pytest
from click.testing import CliRunner

from edgetest.interface import cli

SETUP_CFG = """
[metadata]
name = toy_package
version = 0.1.0
description = Fake description
python_requires =
    >=3.7.0

[options]
zip_safe = False
include_package_data = True
packages = find:
install_requires =
    pandas<=1.0.0

[options.extras_require]
tests =
    pytest

[edgetest]
extras =
    tests
"""

SETUP_PY = """
from setuptools import setup

setup()
"""

MODULE_CODE = """
def main() -> str:
    print("Hello world")
"""

TEST_CODE = """
from toy_package import main

def test_func():
    main()
"""


@pytest.mark.integration
def test_toy_package():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("setup.cfg", "w") as outfile:
            outfile.write(SETUP_CFG)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=setup.cfg"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()
        assert Path(loc, ".edgetest", "pandas").is_dir()



"""Testing the core module."""

import platform
from pathlib import Path
from unittest.mock import PropertyMock, call, patch

import pytest

from edgetest.core import TestPackage


@patch.object(Path, "cwd")
@patch("edgetest.utils.Popen", autospec=True)
def test_basic_setup(mock_popen, mock_path, tmpdir, plugin_manager):
    """Test creating a basic environment."""
    location = tmpdir.mkdir("mydir")
    mock_path.return_value = Path(str(location))
    mock_popen.return_value.communicate.return_value = ("output", "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    tester = TestPackage(
        hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade"]
    )

    assert tester._basedir == Path(str(location)) / ".edgetest"

    tester.setup()

    env_loc = Path(str(location)) / ".edgetest" / "myenv"
    if platform.system() == "Windows":
        py_loc = env_loc / "Scripts" / "python"
    else:
        py_loc = env_loc / "bin" / "python"

    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pip",
                "install",
                "myupgrade",
                "--upgrade",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
    ]


@patch.object(Path, "cwd")
@patch("edgetest.utils.Popen", autospec=True)
def test_basic_setup_error(mock_popen, mock_path, tmpdir, plugin_manager):
    """Test failure in environment creation."""
    location = tmpdir.mkdir("mydir")
    mock_path.return_value = Path(str(location))
    mock_popen.return_value.communicate.return_value = ("output", "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=1)

    tester = TestPackage(
        hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade"]
    )

    assert tester._basedir == Path(str(location)) / ".edgetest"

    with pytest.raises(RuntimeError):
        tester.setup()


@patch.object(Path, "cwd")
@patch("edgetest.utils.Popen", autospec=True)
def test_setup_extras(mock_popen, mock_path, tmpdir, plugin_manager):
    """Test creating an environment with extra installations."""
    location = tmpdir.mkdir("mydir")
    mock_path.return_value = Path(str(location))
    mock_popen.return_value.communicate.return_value = ("output", "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    tester = TestPackage(
        hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade"]
    )

    assert tester._basedir == Path(str(location)) / ".edgetest"

    tester.setup(extras=["tests", "complete"])

    env_loc = str(Path(str(location)) / ".edgetest" / "myenv")
    if platform.system() == "Windows":
        py_loc = Path(env_loc) / "Scripts" / "python"
    else:
        py_loc = Path(env_loc) / "bin" / "python"

    assert mock_popen.call_args_list == [
        call(
            (f"{py_loc}", "-m", "pip", "install", ".[tests, complete]"),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (f"{py_loc}", "-m", "pip", "install", "myupgrade", "--upgrade"),
            stdout=-1,
            universal_newlines=True,
        ),
    ]


@patch.object(Path, "cwd")
@patch("edgetest.utils.Popen", autospec=True)
def test_setup_pip_deps(mock_popen, mock_path, tmpdir, plugin_manager):
    """Test creating an environment with pip dependencies."""
    location = tmpdir.mkdir("mydir")
    mock_path.return_value = Path(str(location))
    mock_popen.return_value.communicate.return_value = ("output", "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    tester = TestPackage(
        hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade"]
    )

    assert tester._basedir == Path(str(location)) / ".edgetest"

    tester.setup(deps=["-r requirements.txt", "otherpkg"])

    env_loc = str(Path(str(location)) / ".edgetest" / "myenv")

    if platform.system() == "Windows":
        py_loc = Path(env_loc) / "Scripts" / "python"
    else:
        py_loc = Path(env_loc) / "bin" / "python"

    assert mock_popen.call_args_list == [
        call(
            (
                f"{py_loc}",
                "-m",
                "pip",
                "install",
                "-r",
                "requirements.txt",
                "otherpkg",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (f"{py_loc}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (f"{py_loc}", "-m", "pip", "install", "myupgrade", "--upgrade"),
            stdout=-1,
            universal_newlines=True,
        ),
    ]


@patch.object(Path, "cwd")
@patch("edgetest.core.Popen", autospec=True)
def test_run_tests(mock_popen, mock_path, tmpdir, plugin_manager):
    """Test running basic tests."""
    location = tmpdir.mkdir("mydir")
    mock_path.return_value = Path(str(location))
    mock_popen.return_value.communicate.return_value = ("output", "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    tester = TestPackage(
        hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade"]
    )

    env_loc = str(Path(str(location)) / ".edgetest" / "myenv")
    if platform.system() == "Windows":
        py_loc = Path(env_loc) / "Scripts" / "python"
    else:
        py_loc = Path(env_loc) / "bin" / "python"

    out = tester.run_tests(command="pytest tests -m 'not integration'")

    assert out == 0
    assert mock_popen.call_args_list == [
        call(
            (f"{py_loc}", "-m", "pytest", "tests", "-m", "not integration"),
            universal_newlines=True,
        )
    ]



"""Core module."""

import json
import shlex
from pathlib import Path
from subprocess import Popen
from typing import Dict, List, Optional

from pluggy._hooks import _HookRelay

from .logger import get_logger
from .utils import _run_command, pushd

LOG = get_logger(__name__)


class TestPackage:
    """Run test commands with bleeding edge dependencies.

    Parameters
    ----------
    hook : _HookRelay
        The hook object from ``pluggy``.
    envname : str
        The name of the conda environment to create/use.
    upgrade : list
        The list of packages to upgrade
    package_dir : str, optional (default None)
        The location of the local package to install and test.

    Attributes
    ----------
    _basedir : str
        The base directory location for each environment
    status : bool
        A boolean status indicator for whether or not the tests passed. Only populated
        after ``run_tests`` has been executed.
    """

    def __init__(
        self,
        hook: _HookRelay,
        envname: str,
        upgrade: List[str],
        package_dir: Optional[str] = None,
    ):
        """Init method."""
        self.hook = hook
        self._basedir = Path(Path.cwd(), ".edgetest")
        self._basedir.mkdir(exist_ok=True)
        self.envname = envname
        self.upgrade = upgrade
        self.package_dir = package_dir or "."

        self.status: bool = False

    @property
    def python_path(self) -> str:
        """Get the path to the python executable.

        Returns
        -------
        str
            The path to the python executable.
        """
        return self.hook.path_to_python(basedir=self._basedir, envname=self.envname)  # type: ignore

    def setup(
        self,
        extras: Optional[List[str]] = None,
        deps: Optional[List[str]] = None,
        **options,
    ) -> None:
        """Set up the testing environment.

        Parameters
        ----------
        extras : list, optional (default None)
            The list of extra installations to include.
        deps : list, optional (default None)
            A list of additional dependencies to install via ``pip``
        **options
            Additional options for ``self.hook.create_environment``.

        Returns
        -------
        None

        Raises
        ------
        RuntimeError
            This error will be raised if any part of the set up process fails.
        """
        # Create the conda environment
        LOG.info(f"Creating the following environment: {self.envname}...")
        self.hook.create_environment(
            basedir=self._basedir, envname=self.envname, conf=options
        )
        LOG.info(f"Successfully created {self.envname}")

        # Install the local package
        with pushd(self.package_dir):
            pkg = "."
            if extras:
                pkg += f"[{', '.join(extras)}]"
            if deps:
                LOG.info(
                    f"Installing specified additional dependencies into {self.envname}..."
                )
                split = [shlex.split(dep) for dep in deps]
                _run_command(
                    self.python_path,
                    "-m",
                    "pip",
                    "install",
                    *[itm for lst in split for itm in lst],
                )
            LOG.info(f"Installing the local package into {self.envname}...")
            _run_command(self.python_path, "-m", "pip", "install", pkg)
            LOG.info(f"Successfully installed the local package into {self.envname}...")

        # Upgrade package(s)
        LOG.info(
            f"Upgrading the following packages in {self.envname}: {', '.join(self.upgrade)}"
        )
        _run_command(
            self.python_path,
            "-m",
            "pip",
            "install",
            *self.upgrade,
            "--upgrade",
        )
        LOG.info(f"Successfully upgraded packages in {self.envname}")

    def upgraded_packages(self) -> List[Dict[str, str]]:
        """Get the list of upgraded packages for the test environment.

        Parameters
        ----------
        None

        Returns
        -------
        List
            The output of ``pip list --format json``, filtered to the packages upgraded
            for this environment.
        """
        # Get the version for the upgraded package(s)
        out, _ = _run_command(self.python_path, "-m", "pip", "list", "--format", "json")
        outjson = json.loads(out)

        return [pkg for pkg in outjson if pkg.get("name", "") in self.upgrade]

    def run_tests(self, command: str) -> int:
        """Run the tests in the package directory.

        Parameters
        ----------
        command : str
            The test command

        Returns
        -------
        int
            The exit code
        """
        with pushd(self.package_dir):
            popen = Popen(
                (self.python_path, "-m", *shlex.split(command)), universal_newlines=True
            )
            popen.communicate()

        self.status = bool(popen.returncode == 0)

        return popen.returncode



"""Generate rST reports."""

from typing import Any, List

from tabulate import tabulate

from .core import TestPackage


def gen_report(testers: List[TestPackage]) -> Any:
    """Generate a rST report.

    Parameters
    ----------
    testers : list
        A list of ``TestPackage`` objects.

    Returns
    -------
    Any
        The report.
    """
    headers = ["Environment", "Passing tests", "Upgraded packages", "Package version"]
    rows: List[List] = []
    for env in testers:
        upgraded = env.upgraded_packages()
        for pkg in upgraded:
            rows.append([env.envname, env.status, pkg["name"], pkg["version"]])

    return tabulate(rows, headers=headers, tablefmt="rst")



"""Hook specifications for edgetest."""

from typing import Dict, List

import pluggy

from .schema import Schema

hookspec = pluggy.HookspecMarker("edgetest")


@hookspec
def addoption(schema: Schema):
    """Modify the schema for custom options.

    You can add environment-level options through ``add_envoption`` or global
    configuration options through ``add_globaloption``.
    """


@hookspec
def pre_run_hook(conf: Dict):
    """Pre-setup and test hook.

    Parameters
    ----------
    conf : Dict
        The entire configuration dictionary.
    """


@hookspec(firstresult=True)
def path_to_python(basedir: str, envname: str) -> str:
    """Return the path to the python executable.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        The name of the virtual environment.

    Returns
    -------
    str
        The path to the python executable for the environment. For installations
        via ``pip``, we'll be running ``python -m pip install ...``, where ``python``
        is the python executable for the environment.
    """


@hookspec(firstresult=True)
def create_environment(basedir: str, envname: str, conf: Dict):
    """Create the virtual environment for testing.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        The name of the virtual environment.
    conf : dict
        The configuration dictionary for the environment. This is useful if you
        want to add configuration arguments for additional dependencies that can
        only be installed through the environment manager (e.g. Conda).

    Raises
    ------
    RuntimeError
        Error raised if the environment cannot be created.
    """


@hookspec
def post_run_hook(testers: List, conf: Dict):
    """Post testing hook.

    For executing code after the environment set up and testing.

    Parameters
    ----------
    testers : list
        The list of ``TestPackage`` objects.
    conf : dict
        The entire configuration dictionary.
    """
